# SPDX-License-Identifier: MIT

import glob
import math
import os
import re
import random
import subprocess
import time
from datetime import timedelta, datetime
from packaging import version
from pyudev import Context

from amd_debug.sleep_report import SleepReport
from amd_debug.database import SleepDatabase
from amd_debug.battery import Batteries
from amd_debug.power_rails import PowerRails
from amd_debug.kernel import get_kernel_log, get_kernel_command_line, sscanf_bios_args
from amd_debug.common import (
    print_color,
    read_file,
    check_lockdown,
    run_countdown,
    BIT,
    AmdTool,
)
from amd_debug.acpi import AcpicaTracer
from amd_debug.failures import (
    AcpiBiosError,
    Irq1Workaround,
    LowHardwareSleepResidency,
    SpuriousWakeup,
    RtcAlarmWrong,
    IommuPageFault,
)


class Headers:
    """Header strings for the debug output"""

    Irq1Workaround = "Disabling IRQ1 wakeup source to avoid platform firmware bug"
    WokeFromIrq = "Woke up from IRQ"
    LastCycleResults = "Results from last s2idle cycle"
    CycleCount = "Suspend cycle"
    SuspendDuration = "Suspend timer programmed for"


def soc_needs_irq1_wa(family, model, smu_version):
    """Check if the SoC needs the IRQ1 workaround"""
    if family == 0x17:
        if model in [0x68, 0x60]:
            return True
    elif family == 0x19:
        if model == 0x50:
            return version.parse(smu_version) < version.parse("64.66.0")
    return False


def toggle_pm_debug(enable):
    """Enable or disable pm_debug_messages"""
    pm_debug_messages = os.path.join("/", "sys", "power", "pm_debug_messages")
    with open(pm_debug_messages, "w", encoding="utf-8") as w:
        w.write("1" if enable else "0")


def pm_debugging(func):
    """Decorator to enable pm_debug_messages"""

    def runner(*args, **kwargs):
        toggle_pm_debug(True)
        ret = func(*args, **kwargs)
        toggle_pm_debug(False)

        return ret

    return runner


class SleepValidator(AmdTool):
    """Class to validate the sleep state"""

    def __init__(self, tool_debug, bios_debug):
        log_prefix = "s2idle" if tool_debug else None
        super().__init__(log_prefix)

        self.pyudev = Context()

        self.kernel_log = get_kernel_log()
        self.db = SleepDatabase()
        self.batteries = Batteries()
        self.power_rails = PowerRails()
        self.acpica = AcpicaTracer()
        self.bios_debug = bios_debug
        self.cpu_family = ""
        self.cpu_model = ""
        self.cpu_model_string = ""
        self.smu_version = ""
        self.smu_program = ""
        self.last_suspend = datetime.now()
        self.requested_duration = 0
        self.userspace_duration = 0
        self.kernel_duration = 0
        self.hw_sleep_duration = 0
        self.failures = []
        self.gpes = {}
        self.display_debug = tool_debug
        self.lockdown = check_lockdown()
        self.logind = False
        self.upep = False
        self.cycle_count = 0
        self.upep = False
        self.upep_microsoft = False
        self.wakeup_irqs = []
        self.idle_masks = []
        self.acpi_errors = []
        self.active_gpios = []
        self.irq1_workaround = False
        self.thermal = {}
        self.wakeup_count = {}
        self.page_faults = []
        self.notify_devices = []

    def capture_running_compositors(self):
        """Capture information about known compositor processes found"""

        known_compositors = [
            "kwin_wayland",
            "gnome-shell",
            "cosmic-session",
            "hyprland",
        ]

        # Get a list of all process directories in /proc
        process_dirs = glob.glob("/proc/[0-9]*")

        # Extract and print the process names
        for proc_dir in process_dirs:
            p = os.path.join(proc_dir, "exe")
            if not os.path.exists(p):
                continue
            exe = os.path.basename(os.readlink(p)).split()[0]
            if exe in known_compositors:
                self.db.record_debug(f"{exe} compositor is running")

    def capture_power_profile(self):
        """Capture power profile information"""
        cmd = ["/usr/bin/powerprofilesctl"]
        if os.path.exists(cmd[0]):
            try:
                output = subprocess.check_output(cmd, stderr=subprocess.STDOUT).decode(
                    "utf-8"
                )
                self.db.record_debug("Power Profiles:")
                lines = output.split("\n")
                lines = [line for line in lines if line.strip()]
                for line in lines:
                    prefix = "│ " if line != lines[-1] else "└─"
                    self.db.record_debug(f"{prefix}{line.strip()}")
            except subprocess.CalledProcessError as e:
                self.db.record_debug("Failed to run powerprofilesctl: %s", e.output)

    def capture_battery(self):
        """Capture battery energy levels"""
        for name in self.batteries.get_batteries():
            energy = self.batteries.get_energy(name)
            full = self.batteries.get_energy_full(name)
            self.db.record_debug(f"{name} energy level is {energy} µWh")
            self.db.record_battery_energy(name, energy, full, "W")

    def capture_power_rails(self):
        """Capture power rail energy levels"""
        if not self.power_rails.rails:
            return  # No IIO power rails detected

        for rail in self.power_rails.rails:
            try:
                energy = rail.read_energy_raw()
                self.db.record_power_rail_energy(
                    rail.label, energy, rail.energy_scale
                )
            except OSError as e:
                self.db.record_debug(
                    f"Failed to read {rail.label}: {e}", priority=1
                )

    def check_rtc_cmos(self):
        """Check if the RTC is configured to use ACPI alarm"""
        p = os.path.join(
            "/", "sys", "module", "rtc_cmos", "parameters", "use_acpi_alarm"
        )
        val = read_file(p)
        if val == "N":
            self.db.record_cycle_data(
                "`rtc_cmos` not configured to use ACPI alarm", "🚦"
            )
            self.failures += [RtcAlarmWrong()]

    def check_gpes(self):
        """Capture general purpose event count"""
        base = os.path.join("/", "sys", "firmware", "acpi", "interrupts")
        for root, _dirs, files in os.walk(base, topdown=False):
            for fname in files:
                if not fname.startswith("gpe") or fname == "gpe_all":
                    continue
                target = os.path.join(root, fname)
                val = 0
                with open(target, "r", encoding="utf-8") as r:
                    val = int(r.read().split()[0])
                if fname in self.gpes and self.gpes[fname] != val:
                    self.db.record_debug(
                        f"{fname} increased from {self.gpes[fname]} to {val}",
                    )
                self.gpes[fname] = val

    def capture_wake_sources(self):
        """Capture possible wakeup sources"""

        def get_input_sibling_name(pyudev, parent):
            """Get the name of the input sibling"""
            for inp in pyudev.list_devices(subsystem="input", parent=parent):
                if not "NAME" in inp.properties:
                    continue
                return inp.properties["NAME"]
            return ""

        devices = []
        for wake_dev in self.pyudev.list_devices(subsystem="wakeup"):
            p = os.path.join(wake_dev.sys_path, "device", "power", "wakeup")
            if not os.path.exists(p):
                continue
            wake_en = read_file(p)
            name = ""
            sys_name = wake_dev.sys_path
            # determine the type of device it hangs off of
            acpi = wake_dev.find_parent(subsystem="acpi")
            serio = wake_dev.find_parent(subsystem="serio")
            rtc = wake_dev.find_parent(subsystem="rtc")
            pci = wake_dev.find_parent(subsystem="pci")
            mhi = wake_dev.find_parent(subsystem="mhi")
            pnp = wake_dev.find_parent(subsystem="pnp")
            hid = wake_dev.find_parent(subsystem="hid")
            thunderbolt_device = wake_dev.find_parent(
                subsystem="thunderbolt", device_type="thunderbolt_device"
            )
            thunderbolt_domain = wake_dev.find_parent(
                subsystem="thunderbolt", device_type="thunderbolt_domain"
            )
            i2c = wake_dev.find_parent(subsystem="i2c")
            if i2c is not None:
                sys_name = i2c.sys_name
                name = get_input_sibling_name(self.pyudev, i2c)
            elif thunderbolt_device is not None:
                if "USB4_TYPE" in thunderbolt_device.properties:
                    name = (
                        f'USB4 {thunderbolt_device.properties["USB4_TYPE"]} controller'
                    )
                sys_name = thunderbolt_device.sys_name
            elif thunderbolt_domain is not None:
                name = "Thunderbolt domain"
                sys_name = thunderbolt_domain.sys_name
            elif serio is not None:
                sys_name = serio.sys_name
                name = get_input_sibling_name(self.pyudev, serio)
            elif rtc is not None:
                sys_name = rtc.sys_name
                for _parent in self.pyudev.list_devices(
                    subsystem="platform", parent=rtc, DRIVER="alarmtimer"
                ):
                    name = "Real Time Clock alarm timer"
                    break
            elif mhi is not None:
                sys_name = mhi.sys_name
                name = "Mobile Broadband host interface"
            elif hid is not None:
                name = hid.properties["HID_NAME"]
                sys_name = hid.sys_name
            elif pci is not None:
                sys_name = pci.sys_name
                if (
                    "ID_PCI_SUBCLASS_FROM_DATABASE" in pci.properties
                    and "ID_VENDOR_FROM_DATABASE" in pci.properties
                ):
                    name = f'{pci.properties["ID_VENDOR_FROM_DATABASE"]} {pci.properties["ID_PCI_SUBCLASS_FROM_DATABASE"]}'
                else:
                    name = f"PCI {pci.properties['PCI_CLASS']}"
            elif acpi is not None:
                sys_name = acpi.sys_name
                if acpi.driver == "button":
                    for inp in self.pyudev.list_devices(subsystem="input", parent=acpi):
                        if not "NAME" in inp.properties:
                            continue
                        name = f"ACPI {inp.properties['NAME']}"
                        break
                elif acpi.driver in ["battery", "ac"]:
                    for ps in self.pyudev.list_devices(
                        subsystem="power_supply", parent=acpi
                    ):
                        if not "POWER_SUPPLY_NAME" in ps.properties:
                            continue
                        name = f"ACPI {ps.properties['POWER_SUPPLY_TYPE']}"
            elif pnp is not None:
                name = "Plug-n-play"
                if pnp.driver == "rtc_cmos":
                    name = f"{name} Real Time Clock"
                sys_name = pnp.sys_name

            name = name.replace('"', "")
            devices.append(f"{name}|{sys_name}|{wake_en}")
        devices.sort()
        debug_str = "Wakeup Source|Linux Device|Status\n"
        for dev in devices:
            debug_str += f"{dev}\n"
        self.db.record_debug(debug_str)

    def capture_lid(self) -> None:
        """Capture lid state"""
        p = os.path.join("/", "proc", "acpi", "button", "lid")
        for root, _dirs, files in os.walk(p):
            for fname in files:
                p = os.path.join(root, fname)
                state = read_file(p).split(":")[1].strip()
                self.db.record_debug(f"ACPI Lid ({p}): {state}")

    def capture_wakeup_irq_data(self) -> bool:
        """Capture the wakeup IRQ to the log"""
        p = os.path.join("/", "sys", "power", "pm_wakeup_irq")
        try:
            n = read_file(p)
            p = os.path.join("/", "sys", "kernel", "irq", n)
            chip_name = read_file(os.path.join(p, "chip_name"))
            name = read_file(os.path.join(p, "name"))
            hw = read_file(os.path.join(p, "hwirq"))
            actions = read_file(os.path.join(p, "actions"))
            message = f"{Headers.WokeFromIrq} {n} ({chip_name} {hw}-{name} {actions})"
            self.db.record_debug(message)
            irq = int(n)
            if irq and irq not in self.wakeup_irqs:
                self.wakeup_irqs += [irq]
        except OSError:
            pass
        return True

    def capture_amdgpu_ips_status(self):
        """Capture the AMDGPU IPS status"""
        for device in self.pyudev.list_devices(subsystem="pci", PCI_CLASS="38000"):
            pci_id = device.properties.get("PCI_ID")
            if not pci_id.startswith("1002"):
                continue
            slot = device.properties.get("PCI_SLOT_NAME")
            p = os.path.join(
                "/", "sys", "kernel", "debug", "dri", slot, "amdgpu_dm_ips_status"
            )
            if not os.path.exists(p):
                continue
            self.db.record_debug("IPS status")
            try:
                lines = read_file(p).split("\n")
                for line in lines:
                    prefix = "│ " if line != lines[-1] else "└─"
                    self.db.record_debug(f"{prefix}{line}")
            except PermissionError:
                if self.lockdown:
                    self.db.record_debug(
                        "Unable to gather IPS state data due to kernel lockdown."
                    )
                else:
                    self.db.record_debug("Failed to read IPS state data")

    def capture_thermal(self):
        """Capture thermal zone information"""
        devs = []
        for dev in self.pyudev.list_devices(subsystem="acpi", DRIVER="thermal"):
            devs.append(dev)
        if not devs:
            return

        self.db.record_debug("Thermal zones")
        for dev in devs:
            prefix = "├─ " if dev != devs[-1] else "└─"
            detail_prefix = "│ \t" if dev != devs[-1] else "  \t"
            name = os.path.basename(dev.device_path)
            p = os.path.join(dev.sys_path, "thermal_zone")
            temp = int(read_file(os.path.join(p, "temp"))) / 1000

            self.db.record_debug(f"{prefix}{name}")
            if name not in self.thermal:
                self.db.record_debug(f"{detail_prefix} temp: {temp}°C")
            else:
                self.db.record_debug(
                    f"{detail_prefix} {self.thermal[name]}°C -> {temp}°C"
                )

            # handle all trip points
            trip_count = 0
            for f in os.listdir(p):
                if "trip_point" not in f:
                    continue
                if "temp" not in f:
                    continue
                trip_count = trip_count + 1

            for i in range(0, trip_count):
                f = os.path.join(p, f"trip_point_{i}_type")
                trip_type = read_file(f)
                f = os.path.join(p, f"trip_point_{i}_temp")
                trip = int(read_file(f)) / 1000

                if name not in self.thermal:
                    self.db.record_debug(f"{detail_prefix} {trip_type} trip: {trip}°C")

                if temp > trip:
                    self.db.record_prereq(
                        f"Thermal zone {name} past trip point {trip_type}: {trip}°C",
                        "🌡️",
                    )
                    return False
            self.thermal[name] = temp

    def capture_input_wakeup_count(self):
        """Capture wakeup count for input related devices"""

        def get_wakeup_count(device):
            """Get the wakeup count for a device"""
            p = os.path.join(device.sys_path, "power", "wakeup")
            if not os.path.exists(p):
                return None
            p = os.path.join(device.sys_path, "power", "wakeup_count")
            if not os.path.exists(p):
                return None
            return read_file(p)

        wakeup_count = {}
        for device in self.pyudev.list_devices(subsystem="input"):
            count = get_wakeup_count(device)
            if count is not None:
                wakeup_count[device.sys_path] = count
                continue
            # iterate parents until finding one with a wakeup count
            # or no more parents
            parent = device.parent
            while parent is not None:
                count = get_wakeup_count(parent)
                if count is not None:
                    wakeup_count[parent.sys_path] = count
                    break
                parent = parent.parent

        # diff the count
        for device, count in wakeup_count.items():
            if device not in self.wakeup_count:
                continue
            if self.wakeup_count[device] == count:
                continue
            self.db.record_debug(
                f"Woke up from input source {device} ({self.wakeup_count[device]}->{count})",
            )
        self.wakeup_count = wakeup_count

    def capture_hw_sleep(self) -> bool:
        """Check for hardware sleep state"""
        # try from kernel 6.4's suspend stats interface first because it works
        # even with kernel lockdown
        p = os.path.join("/", "sys", "power", "suspend_stats", "last_hw_sleep")
        if os.path.exists(p):
            self.hw_sleep_duration = int(read_file(p)) / 10**6
        if not os.path.exists(p) and not self.hw_sleep_duration:
            p = os.path.join("/", "sys", "kernel", "debug", "amd_pmc", "smu_fw_info")
            try:
                val = read_file(p)
                for line in val.split("\n"):
                    if "Last S0i3 Status" in line:
                        continue
                    if "Time (in us) in S0i3" in line:
                        self.hw_sleep_duration = int(line.split(":")[1]) / 10**6
            except PermissionError:
                if self.lockdown:
                    self.db.record_cycle_data(
                        "Unable to gather hardware sleep data with lockdown engaged",
                        "🚦",
                    )
                else:
                    self.db.record_cycle_data(
                        "Failed to read hardware sleep data", "🚦"
                    )
                return False
            except FileNotFoundError:
                self.db.record_debug(f"HW sleep statistics file {p} is missing")
        if not self.hw_sleep_duration:
            self.db.record_cycle_data("Did not reach hardware sleep state", "❌")
        return self.hw_sleep_duration > 0

    def capture_command_line(self):
        """Capture the kernel command line to debug"""
        self.db.record_debug(f"/proc/cmdline: {get_kernel_command_line()}")

    def _analyze_kernel_log_line(self, line, priority):
        bios_args = sscanf_bios_args(line)
        if bios_args:
            if isinstance(bios_args, str):
                line = bios_args
                priority = 7
            else:
                return
        elif "Timekeeping suspended for" in line:
            self.cycle_count += 1
            for f in line.split():
                try:
                    self.kernel_duration += float(f)
                except ValueError:
                    pass
        elif "Successfully transitioned to state" in line:
            self.upep = True
            if "Successfully transitioned to state lps0 ms entry" in line:
                self.upep_microsoft = True
        elif "_DSM function" in line:
            self.upep = True
            if "_DSM function 7" in line:
                self.upep_microsoft = True
        elif "Last suspend in deepest state for" in line:
            for f in line.split():
                if not f.endswith("us"):
                    continue
                try:
                    self.hw_sleep_duration += float(f.strip("us")) / 10**6
                except ValueError:
                    pass
        elif "Triggering wakeup from IRQ" in line:
            irq = int(line.split()[-1])
            if irq and irq not in self.wakeup_irqs:
                self.wakeup_irqs += [irq]
        elif "SMU idlemask s0i3" in line:
            self.idle_masks += [line.split()[-1]]
        elif "ACPI BIOS Error" in line or "ACPI Error" in line:
            self.acpi_errors += [line]
        elif re.search("GPIO.*is active", line):
            self.active_gpios += re.findall(
                r"\d+", re.search("GPIO.*is active", line).group()
            )
        elif Headers.Irq1Workaround in line:
            self.irq1_workaround = True
        # AMD-Vi: Event logged [IO_PAGE_FAULT device=0000:00:0c.0 domain=0x0000 address=0x7e800000 flags=0x0050]
        elif "Event logged [IO_PAGE_FAULT" in line:
            # get the device from string
            device = re.search(r"device=(.*?) domain", line)
            if device:
                device = device.group(1)
                if device not in self.page_faults:
                    self.page_faults += [device]

        # evmisc-0132 ev_queue_notify_reques: Dispatching Notify on [UBTC] (Device) Value 0x80 (Status Change) Node 0000000080144eee
        if "Dispatching Notify on" in line:
            # add device without the [] to notify_devices if it's not already there
            device = re.search(r"\[(.*?)\]", line)
            if device:
                device = device.group(1)
                if device not in self.notify_devices:
                    self.notify_devices += [device]
            priority = 7

        self.db.record_debug(line, priority)

    def analyze_kernel_log(self):
        """Analyze one of the lines from the kernel log"""
        self.kernel_log.process_callback(self._analyze_kernel_log_line)

        if self.cycle_count:
            self.db.record_cycle_data(
                f"Hardware sleep cycle count: {self.cycle_count}",
                "💤",
            )
        if self.wakeup_irqs:
            if 1 in self.wakeup_irqs and soc_needs_irq1_wa(
                self.cpu_family, self.cpu_model, self.smu_version
            ):
                if self.irq1_workaround:
                    self.db.record_cycle_data(
                        "Kernel workaround for IRQ1 issue utilized", "○"
                    )
                else:
                    self.db.record_cycle_data("IRQ1 found during wakeup", "🚦")
                    self.failures += [Irq1Workaround()]
        if self.idle_masks:
            bit_changed = 0
            for i, mask_i in enumerate(self.idle_masks):
                for _j, mask_j in enumerate(self.idle_masks[i:], start=i):
                    if mask_i != mask_j:
                        bit_changed = bit_changed | (int(mask_i, 16) & ~int(mask_j, 16))
            if bit_changed:
                for bit in range(0, 31):
                    if bit_changed & BIT(bit):
                        self.db.record_debug(
                            f"Idle mask bit {bit} (0x{BIT(bit):x}) changed during suspend",
                        )
        if self.upep:
            if self.upep_microsoft:
                self.db.record_debug("Used Microsoft uPEP GUID in LPS0 _DSM")
            else:
                self.db.record_debug("Used AMD uPEP GUID in LPS0 _DSM")
        if self.acpi_errors:
            self.db.record_cycle_data("ACPI BIOS errors found", "❌")
            self.failures += [AcpiBiosError(self.acpi_errors)]
        if self.page_faults:
            self.db.record_cycle_data("Page faults found", "❌")
            self.failures += [IommuPageFault(self.page_faults)]
        if self.notify_devices:
            self.db.record_cycle_data(
                f"Notify devices {self.notify_devices} found during suspend", "💤"
            )

    def analyze_duration(self, t0, t1, requested, kernel, hw):
        """Analyze the duration of the last cycle"""
        userspace_duration = t1 - t0
        min_suspend_duration = timedelta(seconds=requested * 0.9)
        expected_wake_time = t0 + min_suspend_duration
        if t1 > expected_wake_time:
            print_color(
                f"Userspace suspended for {userspace_duration}",
                "✅",
            )
        else:
            print_color(
                f"Userspace suspended for {userspace_duration} (< minimum expected {min_suspend_duration})",
                "❌",
            )
            self.failures += [SpuriousWakeup(requested, userspace_duration)]
        percent = float(kernel) / userspace_duration.total_seconds()
        print_color(
            f"Kernel suspended for total of {timedelta(seconds=kernel)} ({percent:.2%})",
            "✅",
        )

        percent = float(hw / userspace_duration.total_seconds())
        if userspace_duration.total_seconds() >= 60:
            if percent > 0.9:
                symbol = "✅"
            else:
                symbol = "❌"
                self.failures += [
                    LowHardwareSleepResidency(userspace_duration, percent)
                ]
        else:
            symbol = "✅"
        print_color(
            "In a hardware sleep state for {time} {percent_msg}".format(
                time=timedelta(seconds=hw),
                percent_msg="" if not percent else "({:.2%})".format(percent),
            ),
            symbol,
        )

    def post(self):
        """Post-process the suspend test results"""
        self.cycle_count = 0
        self.upep = False
        self.upep_microsoft = False
        self.wakeup_irqs = []
        self.idle_masks = []
        self.acpi_errors = []
        self.active_gpios = []
        self.notify_devices = []
        self.page_faults = []
        self.irq1_workaround = False

        checks = [
            self.capture_wakeup_irq_data,
            self.analyze_kernel_log,
            self.check_gpes,
            self.capture_lid,
            self.check_rtc_cmos,
            self.capture_hw_sleep,
            self.capture_battery,
            self.capture_power_rails,
            self.capture_amdgpu_ips_status,
            self.capture_thermal,
            self.capture_input_wakeup_count,
            self.acpica.restore,
        ]
        for check in checks:
            check()
        self.db.record_cycle(
            self.requested_duration,
            ",".join(str(gpio) for gpio in self.active_gpios),
            ",".join(str(irq) for irq in self.wakeup_irqs),
            int(self.kernel_duration),
            int(self.hw_sleep_duration),
        )

    def prep(self):
        """Prepare the system for suspend testing"""
        self.last_suspend = datetime.now()
        self.kernel_log.seek_tail(self.last_suspend)
        self.db.start_cycle(self.last_suspend)
        self.kernel_duration = 0
        self.hw_sleep_duration = 0
        self.capture_battery()
        self.capture_power_rails()
        self.check_gpes()
        self.capture_lid()
        self.capture_command_line()
        self.capture_wake_sources()
        self.capture_running_compositors()
        self.capture_power_profile()
        self.capture_amdgpu_ips_status()
        self.capture_thermal()
        self.capture_input_wakeup_count()
        if self.bios_debug:
            self.acpica.trace_bios()
        else:
            self.acpica.trace_notify()
        self.db.record_cycle()

    def program_wakealarm(self):
        """Program the RTC wakealarm to wake the system after the requested duration"""
        wakealarm = None
        for device in self.pyudev.list_devices(subsystem="rtc"):
            target = os.path.join(device.sys_path, "wakealarm")
            if os.path.exists(target):
                wakealarm = target
                break
        if wakealarm:
            with open(wakealarm, "w", encoding="utf-8") as w:
                w.write("0")
            with open(wakealarm, "w", encoding="utf-8") as w:
                w.write("+%s\n" % self.requested_duration)
        else:
            print_color("No RTC device found, please manually wake system", "🚦")

    def toggle_nvidia(self, value):
        """Write to the NVIDIA suspend interface"""
        p = os.path.join("/", "proc", "driver", "nvidia", "suspend")
        if not os.path.exists(p):
            return True
        fd = os.open(p, os.O_WRONLY | os.O_SYNC)
        try:
            os.write(fd, value)
        except OSError as e:
            self.db.record_cycle_data(f"Failed to set {value} in NVIDIA {e}", "❌")
            return False
        finally:
            os.close(fd)
        self.db.record_debug(f"Wrote {value} to NVIDIA driver")
        return True

    @pm_debugging
    def suspend_system(self):
        """Suspend the system using the dbus or sysfs interface"""

        def get_wakeup_count():
            """Get the wakeup count"""
            p = os.path.join("/", "sys", "power", "wakeup_count")
            if not os.path.exists(p):
                return 0
            try:
                with open(p, "r", encoding="utf-8") as r:
                    return int(r.read())
            except OSError:
                return 0

        if self.logind:
            try:
                import dbus

                bus = dbus.SystemBus()
                obj = bus.get_object(
                    "org.freedesktop.login1", "/org/freedesktop/login1"
                )
                intf = dbus.Interface(obj, "org.freedesktop.login1.Manager")
                propf = dbus.Interface(obj, "org.freedesktop.DBus.Properties")
                if intf.CanSuspend() != "yes":
                    self.db.record_cycle_data("Unable to suspend", "❌")
                    return False
                intf.Suspend(True)
                while propf.Get("org.freedesktop.login1.Manager", "PreparingForSleep"):
                    time.sleep(1)
                return True
            except dbus.exceptions.DBusException as e:
                self.db.record_cycle_data(
                    f"Unable to communicate with logind: {e}", "❌"
                )
                return False
            except ImportError:
                self.db.record_cycle_data("Missing dbus", "❌")
                return False
        else:
            if not self.toggle_nvidia(b"suspend"):
                return False
            old = get_wakeup_count()
            p = os.path.join("/", "sys", "power", "state")
            fd = os.open(p, os.O_WRONLY | os.O_SYNC)
            try:
                os.write(fd, b"mem")
            except OSError as e:
                new = get_wakeup_count()
                self.db.record_cycle_data(
                    f"Failed to set suspend state ({old} -> {new}): {e}", "❌"
                )
                return False
            finally:
                os.close(fd)
            if not self.toggle_nvidia(b"resume"):
                return False
            return True

    def unlock_session(self):
        """Unlock the session using logind"""
        if self.logind:
            try:
                import dbus

                bus = dbus.SystemBus()
                obj = bus.get_object(
                    "org.freedesktop.login1", "/org/freedesktop/login1"
                )
                intf = dbus.Interface(obj, "org.freedesktop.login1.Manager")
                intf.UnlockSessions()
            except dbus.exceptions.DBusException as e:
                self.db.record_cycle_data(
                    f"Unable to communicate with logind: {e}", "❌"
                )
                return False
        return True

    def run(self, duration, count, wait, rand, logind):
        """Run the suspend test"""
        min_duration = 4
        if not count:
            return True

        if logind:
            self.logind = True

        if rand:
            if duration <= min_duration:
                print_color(f"Invalid max duration {duration}", "❌")
                self.db.sync()
                self.report_cycle()
                return False
            print_color(
                f"Running {count} cycle random test with min duration of {min_duration}s, max duration of {duration}s and a max wait of {wait}s",
                "🗣️",
            )
        elif count > 1:
            length = timedelta(seconds=(duration + wait) * count)
            print_color(
                f"Running {count} cycles (Test finish expected @ {datetime.now() + length})".format(),
                "🗣️",
            )
        for i in range(1, count + 1):
            if rand:
                self.requested_duration = random.randint(min_duration, duration)
                requested_wait = random.randint(1, wait)
            else:
                self.requested_duration = duration
                requested_wait = wait
            run_countdown("Suspending system", math.ceil(requested_wait / 2))
            self.prep()
            self.db.record_debug(
                f"{Headers.SuspendDuration} {timedelta(seconds=self.requested_duration)}",
            )
            if count > 1:
                header = f"{Headers.CycleCount} {i}: "
            else:
                header = ""
            print_color(
                f"{header}Started at {self.last_suspend} (cycle finish expected @ {datetime.now() + timedelta(seconds=self.requested_duration + requested_wait)})",
                "🗣️",
            )
            self.program_wakealarm()
            if not self.suspend_system():
                self.db.sync()
                self.report_cycle()
                return False
            run_countdown("Collecting data", math.ceil(requested_wait / 2))
            self.post()
            self.db.sync()
            self.report_cycle()
        self.unlock_session()
        return True

    def systemd_pre_hook(self):
        """Called before suspend"""
        self.prep()
        self.db.sync()
        toggle_pm_debug(True)

    def systemd_post_hook(self):
        """Called after resume"""
        toggle_pm_debug(False)
        t0 = self.db.get_last_cycle()
        self.last_suspend = datetime.strptime(str(t0[0]), "%Y%m%d%H%M%S")
        self.kernel_log.seek_tail(self.last_suspend)
        self.db.start_cycle(self.last_suspend)
        self.post()
        self.db.sync()

    def report_cycle(self):
        """Report the results of the last cycle"""
        print_color(Headers.LastCycleResults, "🗣️")

        app = SleepReport(
            since=self.last_suspend,
            until=self.last_suspend,
            fname=None,
            fmt="stdout",
            tool_debug=self.display_debug,
            report_debug=False,
        )
        app.run(inc_prereq=False)
        return



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the launcher in the amd-debug-tools package.
"""
from unittest.mock import patch

import logging
import unittest

import amd_debug


class TestLauncher(unittest.TestCase):
    """Test launcher functions"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    def setUp(self):
        launcher = None

    def test_launcher_unknown(self):
        """Test launching as unknown exe"""

        with patch("builtins.print") as mock_print:
            result = amd_debug.launch_tool("unknown_exe.py")
            mock_print.assert_called_once_with(
                "\033[91mUnknown exe: unknown_exe.py\033[0m"
            )
            self.assertIsNotNone(result)

    def test_launcher_amd_s2idle(self):
        """Test launching amd_s2idle"""

        with patch("amd_debug.s2idle.main") as mock_main:
            amd_debug.launch_tool("amd_s2idle.py")
            mock_main.assert_called_once()

    def test_launcher_amd_bios(self):
        """Test launching amd_bios"""

        with patch("amd_debug.bios.main") as mock_main:
            amd_debug.launch_tool("amd_bios.py")
            mock_main.assert_called_once()

    def test_launcher_amd_pstate(self):
        """Test launching amd_pstate"""

        with patch("amd_debug.pstate.main") as mock_main:
            amd_debug.launch_tool("amd_pstate.py")
            mock_main.assert_called_once()

    def test_launcher_amd_ttm(self):
        """Test launching amd_ttm"""

        with patch("amd_debug.ttm.main") as mock_main:
            amd_debug.launch_tool("amd_ttm.py")
            mock_main.assert_called_once()



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the bios tool in the amd-debug-tools package.
"""
import argparse
import logging
import unittest
from unittest.mock import patch, MagicMock

from amd_debug.bios import AmdBios, parse_args, main


class TestAmdBios(unittest.TestCase):
    """Test AmdBios class"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    @patch("amd_debug.bios.get_kernel_log")
    def test_init(self, mock_get_kernel_log):
        """Test initialization of AmdBios class"""
        mock_kernel_log = MagicMock()
        mock_get_kernel_log.return_value = mock_kernel_log

        app = AmdBios("test_input", debug=True)

        self.assertEqual(app.kernel_log, mock_kernel_log)

    @patch("amd_debug.bios.relaunch_sudo")
    @patch("amd_debug.bios.minimum_kernel")
    @patch("amd_debug.bios.AcpicaTracer")
    @patch("amd_debug.bios.print_color")
    @patch("subprocess.run")
    def test_set_tracing_enable(
        self,
        _mock_run,
        _mock_print,
        mock_acpica_tracer,
        mock_minimum_kernel,
        mock_relaunch_sudo,
    ):
        """Test enabling tracing"""
        mock_minimum_kernel.return_value = True
        mock_tracer = MagicMock()
        mock_acpica_tracer.return_value = mock_tracer
        mock_tracer.trace_bios.return_value = True

        app = AmdBios(None, debug=False)
        result = app.set_tracing(enable=True)

        mock_relaunch_sudo.assert_called_once()
        mock_tracer.trace_bios.assert_called_once()
        self.assertTrue(result)

    @patch("amd_debug.bios.relaunch_sudo")
    @patch("amd_debug.bios.minimum_kernel")
    @patch("amd_debug.bios.AcpicaTracer")
    @patch("amd_debug.bios.print_color")
    @patch("subprocess.run")
    def test_set_tracing_disable(
        self,
        _mock_run,
        _mock_print,
        mock_acpica_tracer,
        mock_minimum_kernel,
        mock_relaunch_sudo,
    ):
        """Test disabling tracing"""
        mock_minimum_kernel.return_value = True
        mock_tracer = MagicMock()
        mock_acpica_tracer.return_value = mock_tracer
        mock_tracer.disable.return_value = True

        app = AmdBios(None, debug=False)
        result = app.set_tracing(enable=False)

        mock_relaunch_sudo.assert_called_once()
        mock_tracer.disable.assert_called_once()
        self.assertTrue(result)

    @patch("amd_debug.bios.sscanf_bios_args")
    @patch("amd_debug.bios.print_color")
    @patch("subprocess.run")
    def test_analyze_kernel_log_line(
        self, _mock_run, mock_print_color, mock_sscanf_bios_args
    ):
        """Test analyzing kernel log line"""
        mock_sscanf_bios_args.return_value = "BIOS argument found"

        app = AmdBios(None, debug=False)
        app._analyze_kernel_log_line(  # pylint: disable=protected-access
            "test log line", priority="INFO"
        )

        mock_sscanf_bios_args.assert_called_once_with("test log line")
        mock_print_color.assert_called_once_with("BIOS argument found", "🖴")

    @patch("amd_debug.bios.sscanf_bios_args")
    @patch("amd_debug.bios.print_color")
    @patch("subprocess.run")
    def test_analyze_kernel_log_line_no_bios_args(
        self, _mock_run, mock_print_color, mock_sscanf_bios_args
    ):
        """Test analyzing kernel log line with no BIOS arguments"""
        mock_sscanf_bios_args.return_value = None

        app = AmdBios(None, debug=False)
        app._analyze_kernel_log_line(  # pylint: disable=protected-access
            "[123.456] test log line", priority="INFO"
        )

        mock_sscanf_bios_args.assert_called_once_with("[123.456] test log line")
        mock_print_color.assert_called_once_with("test log line", "INFO")

    @patch("amd_debug.bios.get_kernel_log")
    def test_run(self, _mock_run):
        """Test run method"""
        mock_kernel_log = MagicMock()
        app = AmdBios(None, debug=False)
        app.kernel_log = mock_kernel_log

        app.run()

        mock_kernel_log.process_callback.assert_called_once_with(
            app._analyze_kernel_log_line  # pylint: disable=protected-access
        )

    @patch("sys.argv", ["bios.py", "parse", "--input", "test.log", "--tool-debug"])
    def test_parse_args_parse_command(self):
        """Test parse_args with parse command"""

        args = parse_args()
        self.assertEqual(args.command, "parse")
        self.assertEqual(args.input, "test.log")
        self.assertTrue(args.tool_debug)

    @patch("sys.argv", ["bios.py", "trace", "--enable", "--tool-debug"])
    def test_parse_args_trace_enable(self):
        """Test parse_args with trace enable command"""

        args = parse_args()
        self.assertEqual(args.command, "trace")
        self.assertTrue(args.enable)
        self.assertFalse(args.disable)
        self.assertTrue(args.tool_debug)

    @patch("sys.argv", ["bios.py", "trace", "--disable"])
    def test_parse_args_trace_disable(self):
        """Test parse_args with trace disable command"""

        args = parse_args()
        self.assertEqual(args.command, "trace")
        self.assertFalse(args.enable)
        self.assertTrue(args.disable)

    @patch("sys.argv", ["bios.py", "--version"])
    def test_parse_args_version_command(self):
        """Test parse_args with version command"""

        args = parse_args()
        self.assertTrue(args.version)

    @patch("sys.argv", ["bios.py"])
    @patch("argparse.ArgumentParser.print_help")
    @patch("sys.exit")
    def test_parse_args_no_arguments(self, mock_exit, mock_print_help):
        """Test parse_args with no arguments"""

        parse_args()
        mock_print_help.assert_called_once()
        mock_exit.assert_called_once_with(1)

    @patch("sys.argv", ["bios.py", "trace", "--enable", "--disable"])
    @patch("sys.exit")
    def test_parse_args_conflicting_trace_arguments(self, mock_exit):
        """Test parse_args with conflicting trace arguments"""

        parse_args()
        mock_exit.assert_called_once_with("can't set both enable and disable")

    @patch("sys.argv", ["bios.py", "trace"])
    @patch("sys.exit")
    def test_parse_args_missing_trace_arguments(self, mock_exit):
        """Test parse_args with missing trace arguments"""

        parse_args()
        mock_exit.assert_called_once_with("must set either enable or disable")

    @patch("amd_debug.bios.AmdBios")
    @patch("amd_debug.bios.parse_args")
    @patch("amd_debug.bios.version")
    @patch("amd_debug.bios.show_log_info")
    def test_main_trace_command(
        self, mock_show_log_info, _mock_version, mock_parse_args, mock_amd_bios
    ):
        """Test main function with trace command"""
        mock_app = MagicMock()
        mock_amd_bios.return_value = mock_app
        mock_parse_args.return_value = argparse.Namespace(
            command="trace", enable=True, disable=False, tool_debug=True
        )
        mock_app.set_tracing.return_value = True

        result = main()

        mock_parse_args.assert_called_once()
        mock_amd_bios.assert_called_once_with(None, True)
        mock_app.set_tracing.assert_called_once_with(True)
        mock_show_log_info.assert_called_once()
        self.assertIsNone(result)

    @patch("amd_debug.bios.AmdBios")
    @patch("amd_debug.bios.parse_args")
    @patch("amd_debug.bios.version")
    @patch("amd_debug.bios.show_log_info")
    def test_main_parse_command(
        self, mock_show_log_info, _mock_version, mock_parse_args, mock_amd_bios
    ):
        """Test main function with parse command"""
        mock_app = MagicMock()
        mock_amd_bios.return_value = mock_app
        mock_parse_args.return_value = argparse.Namespace(
            command="parse", input="test.log", tool_debug=True
        )
        mock_app.run.return_value = True

        result = main()

        mock_parse_args.assert_called_once()
        mock_amd_bios.assert_called_once_with("test.log", True)
        mock_app.run.assert_called_once()
        mock_show_log_info.assert_called_once()
        self.assertIsNone(result)

    @patch("amd_debug.bios.parse_args")
    @patch("amd_debug.bios.version")
    @patch("amd_debug.bios.show_log_info")
    @patch("amd_debug.bios.print")
    def test_main_version_command(
        self, _mock_print, mock_show_log_info, mock_version, mock_parse_args
    ):
        """Test main function with version command"""
        mock_parse_args.return_value = argparse.Namespace(version=True, command=None)
        mock_version.return_value = "1.0.0"

        result = main()

        mock_parse_args.assert_called_once()
        mock_version.assert_called_once()
        mock_show_log_info.assert_called_once()
        self.assertEqual(result, 1)

    @patch("amd_debug.bios.parse_args")
    @patch("amd_debug.bios.show_log_info")
    def test_main_invalid_command(self, mock_show_log_info, mock_parse_args):
        """Test main function with an invalid command"""
        mock_parse_args.return_value = argparse.Namespace(
            version=False, command="invalid"
        )

        result = main()

        mock_parse_args.assert_called_once()
        mock_show_log_info.assert_called_once()
        self.assertEqual(result, 1)



# SPDX-License-Identifier: MIT
from pyudev import Context


def _get_property(prop, key, fallback="") -> str:
    """Get the property from the given key"""
    try:
        return prop.get(key, fallback)
    except UnicodeDecodeError:
        return ""


class Batteries:
    def __init__(self) -> None:
        self.pyudev = Context()

    def _get_battery(self, name) -> object:
        """Get the battery device object for the given name"""
        for dev in self.pyudev.list_devices(
            subsystem="power_supply", POWER_SUPPLY_TYPE="Battery"
        ):
            if not "PNP0C0A" in dev.device_path:
                continue
            desc = _get_property(dev.properties, "POWER_SUPPLY_NAME", "Unknown")
            if desc != name:
                continue
            return dev
        return None

    def get_batteries(self) -> list:
        """Get a list of battery names on the system"""
        names = []
        for dev in self.pyudev.list_devices(
            subsystem="power_supply", POWER_SUPPLY_TYPE="Battery"
        ):
            if not "PNP0C0A" in dev.device_path:
                continue
            names.append(_get_property(dev.properties, "POWER_SUPPLY_NAME", "Unknown"))
        return names

    def _get_design_voltage(self, dev) -> str:
        """Get the design voltage in µV for the given battery"""
        for key in [
            "POWER_SUPPLY_VOLTAGE_MAX_DESIGN",
            "POWER_SUPPLY_VOLTAGE_MIN_DESIGN",
            "POWER_SUPPLY_VOLTAGE_NOW",
        ]:
            voltage = _get_property(dev.properties, key)
            if voltage:
                return voltage
        return ""

    def _charge_to_energy(self, dev, charge):
        """Convert battery charge in µAh to energy in µWh"""
        voltage = self._get_design_voltage(dev)
        if not charge or not voltage:
            return ""
        return str(round(int(charge) * int(voltage) / 1000000))

    def get_energy(self, name) -> int:
        """Get the current energy for the given battery name"""
        dev = self._get_battery(name)
        if not dev:
            return ""
        energy = _get_property(dev.properties, "POWER_SUPPLY_ENERGY_NOW")
        if not energy:
            charge = _get_property(dev.properties, "POWER_SUPPLY_CHARGE_NOW")
            energy = self._charge_to_energy(dev, charge)
        return energy

    def get_energy_full(self, name) -> int:
        """Get the energy when full for the given battery name"""
        dev = self._get_battery(name)
        if not dev:
            return ""
        energy = _get_property(dev.properties, "POWER_SUPPLY_ENERGY_FULL")
        if not energy:
            charge = _get_property(dev.properties, "POWER_SUPPLY_CHARGE_FULL")
            energy = self._charge_to_energy(dev, charge)
        return energy

    def get_description_string(self, name) -> str:
        """Get a description string for the given battery name"""
        dev = self._get_battery(name)
        if not dev:
            return ""
        man = _get_property(dev.properties, "POWER_SUPPLY_MANUFACTURER", "")
        model = _get_property(dev.properties, "POWER_SUPPLY_MODEL_NAME", "")
        full = self.get_energy_full(name)
        full_design = _get_property(dev.properties, "POWER_SUPPLY_ENERGY_FULL_DESIGN")
        if not full_design:
            charge_full_design = _get_property(
                dev.properties, "POWER_SUPPLY_CHARGE_FULL_DESIGN"
            )
            full_design = self._charge_to_energy(dev, charge_full_design)

        percent = float(full) / int(full_design)
        return f"Battery {name} ({man} {model}) is operating at {percent:.2%} of design"



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the ttm tool in the amd-debug-tools package.
"""
import unittest
import sys
import logging
from unittest import mock

from amd_debug.ttm import main, parse_args, AmdTtmTool, maybe_reboot


class TestParseArgs(unittest.TestCase):
    """Test parse_args function"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    def setUp(self):
        self.default_sys_argv = sys.argv

    def tearDown(self):
        sys.argv = self.default_sys_argv

    @mock.patch.object(sys, "argv", new=["ttm", "--version"])
    def test_parse_args_version(self):
        """Test version argument"""
        args = parse_args()
        self.assertTrue(args.version)
        self.assertFalse(args.set)
        self.assertFalse(args.clear)

    @mock.patch.object(sys, "argv", new=["ttm", "--clear"])
    def test_parse_args_clear(self):
        """Test clear argument"""
        args = parse_args()
        self.assertFalse(args.version)
        self.assertFalse(args.set)
        self.assertTrue(args.clear)


class TestMainFunction(unittest.TestCase):
    """Test main() function logic"""

    @mock.patch("amd_debug.ttm.parse_args")
    @mock.patch("amd_debug.ttm.version", return_value="1.2.3")
    @mock.patch("builtins.print")
    def test_main_version(self, mock_print, _mock_version, mock_parse_args):
        """Test main function with version argument"""
        mock_parse_args.return_value = mock.Mock(
            version=True, set=None, clear=False, tool_debug=False
        )
        ret = main()
        mock_print.assert_called_with("1.2.3")
        self.assertIsNone(ret)

    @mock.patch("amd_debug.ttm.parse_args")
    @mock.patch("amd_debug.ttm.AmdTtmTool")
    @mock.patch("builtins.print")
    def test_main_set_invalid(self, mock_print, _mock_tool, mock_parse_args):
        """Test main function with invalid set argument"""
        mock_parse_args.return_value = mock.Mock(
            version=False, set=0, clear=False, tool_debug=False
        )
        ret = main()
        mock_print.assert_called_with("Error: GB value must be greater than 0")
        self.assertEqual(ret, 1)

    @mock.patch("amd_debug.ttm.parse_args")
    @mock.patch("amd_debug.ttm.AmdTtmTool")
    def test_main_set_valid(self, mock_tool, mock_parse_args):
        """Test main function with set argument"""
        instance = mock_tool.return_value
        instance.set.return_value = True
        mock_parse_args.return_value = mock.Mock(
            version=False, set=2, clear=False, tool_debug=False
        )
        ret = main()
        instance.set.assert_called_with(2)
        self.assertIsNone(ret)

    @mock.patch("amd_debug.ttm.parse_args")
    @mock.patch("amd_debug.ttm.AmdTtmTool")
    def test_main_set_failed(self, mock_tool, mock_parse_args):
        instance = mock_tool.return_value
        instance.set.return_value = False
        mock_parse_args.return_value = mock.Mock(
            version=False, set=2, clear=False, tool_debug=False
        )
        ret = main()
        instance.set.assert_called_with(2)
        self.assertEqual(ret, 1)

    @mock.patch("amd_debug.ttm.parse_args")
    @mock.patch("amd_debug.ttm.AmdTtmTool")
    def test_main_clear_success(self, mock_tool, mock_parse_args):
        """Test main function with clear argument"""
        instance = mock_tool.return_value
        instance.clear.return_value = True
        mock_parse_args.return_value = mock.Mock(
            version=False, set=None, clear=True, tool_debug=False
        )
        ret = main()
        instance.clear.assert_called_once()
        self.assertIsNone(ret)

    @mock.patch("amd_debug.ttm.parse_args")
    @mock.patch("amd_debug.ttm.AmdTtmTool")
    def test_main_clear_failed(self, mock_tool, mock_parse_args):
        """Test main function with clear argument failure"""
        instance = mock_tool.return_value
        instance.clear.return_value = False
        mock_parse_args.return_value = mock.Mock(
            version=False, set=None, clear=True, tool_debug=False
        )
        ret = main()
        instance.clear.assert_called_once()
        self.assertEqual(ret, 1)

    @mock.patch("amd_debug.ttm.parse_args")
    @mock.patch("amd_debug.ttm.AmdTtmTool")
    def test_main_get_success(self, mock_tool, mock_parse_args):
        """Test main function with get argument"""
        instance = mock_tool.return_value
        instance.get.return_value = True
        mock_parse_args.return_value = mock.Mock(
            version=False, set=None, clear=False, tool_debug=False
        )
        ret = main()
        instance.get.assert_called_once()
        self.assertIsNone(ret)

    @mock.patch("amd_debug.ttm.parse_args")
    @mock.patch("amd_debug.ttm.AmdTtmTool")
    def test_main_get_failed(self, mock_tool, mock_parse_args):
        """Test main function with get argument failure"""
        instance = mock_tool.return_value
        instance.get.return_value = False
        mock_parse_args.return_value = mock.Mock(
            version=False, set=None, clear=False, tool_debug=False
        )
        ret = main()
        instance.get.assert_called_once()
        self.assertEqual(ret, 1)


class TestMaybeReboot(unittest.TestCase):
    """Test maybe_reboot function"""

    @mock.patch("builtins.input", return_value="y")
    @mock.patch("amd_debug.ttm.reboot", return_value=True)
    def test_maybe_reboot_yes(self, mock_reboot, _mock_input):
        """Test reboot confirmation and execution"""
        result = maybe_reboot()
        mock_reboot.assert_called_once()
        self.assertTrue(result)

    @mock.patch("builtins.input", return_value="n")
    @mock.patch("amd_debug.ttm.reboot", return_value=True)
    def test_maybe_reboot_no(self, mock_reboot, _mock_input):
        """Test reboot confirmation without execution"""
        result = maybe_reboot()
        mock_reboot.assert_not_called()
        self.assertTrue(result)


class TestAmdTtmTool(unittest.TestCase):
    """Unit tests for AmdTtmTool class"""

    def setUp(self):
        self.tool = AmdTtmTool(logging=False)

    @mock.patch("builtins.open", new_callable=mock.mock_open, read_data="4096")
    @mock.patch("amd_debug.ttm.bytes_to_gb", return_value=4.0)
    @mock.patch("amd_debug.ttm.print_color")
    @mock.patch("amd_debug.ttm.get_system_mem", return_value=16.0)
    def test_get_success(self, _mock_mem, mock_print, _mock_bytes_to_gb, mock_open):
        """Test get() when TTM_PARAM_PATH exists"""
        result = self.tool.get()
        mock_open.assert_called_once_with(
            "/sys/module/ttm/parameters/pages_limit", "r", encoding="utf-8"
        )
        mock_print.assert_any_call(
            "Current TTM pages limit: 4096 pages (4.00 GB)", "💻"
        )
        mock_print.assert_any_call("Total system memory: 16.00 GB", "💻")
        self.assertTrue(result)

    @mock.patch("builtins.open", side_effect=FileNotFoundError)
    @mock.patch("amd_debug.ttm.print_color")
    def test_get_file_not_found(self, mock_print, _mock_open):
        """Test get() when TTM_PARAM_PATH does not exist"""
        result = self.tool.get()
        mock_print.assert_called_with(
            "Error: Could not find /sys/module/ttm/parameters/pages_limit", "❌"
        )
        self.assertFalse(result)

    @mock.patch("amd_debug.ttm.relaunch_sudo", return_value=True)
    @mock.patch("amd_debug.ttm.get_system_mem", return_value=8.0)
    @mock.patch("amd_debug.ttm.print_color")
    def test_set_gb_greater_than_total(
        self, mock_print, _mock_mem, _mock_relaunch_sudo
    ):
        """Test set() when gb_value > total system memory"""
        result = self.tool.set(16)
        mock_print.assert_any_call(
            "16.00 GB is greater than total system memory (8.00 GB)", "❌"
        )
        self.assertFalse(result)

    @mock.patch("amd_debug.ttm.relaunch_sudo", return_value=True)
    @mock.patch("amd_debug.ttm.get_system_mem", return_value=10.0)
    @mock.patch("amd_debug.ttm.print_color")
    @mock.patch("builtins.input", return_value="n")
    def test_set_gb_exceeds_max_percentage_cancel(
        self, _mock_input, mock_print, _mock_mem, mock_relaunch_sudo
    ):
        """Test set() when gb_value exceeds max percentage and user cancels"""
        result = self.tool.set(9.5)
        self.assertFalse(result)
        mock_print.assert_any_call("Operation cancelled.", "🚦")

    @mock.patch("amd_debug.ttm.relaunch_sudo", return_value=True)
    @mock.patch("amd_debug.ttm.get_system_mem", return_value=10.0)
    @mock.patch("amd_debug.ttm.gb_to_pages", return_value=20480)
    @mock.patch("amd_debug.ttm.print_color")
    @mock.patch("builtins.open", new_callable=mock.mock_open)
    @mock.patch("builtins.input", return_value="y")
    @mock.patch("amd_debug.ttm.maybe_reboot", return_value=True)
    def test_set_success(
        self,
        _mock_reboot,
        _mock_input,
        mock_open,
        mock_print,
        _mock_gb_to_pages,
        _mock_mem,
        _relaunch_sudo,
    ):
        """Test set() success path"""
        result = self.tool.set(5)
        mock_open.assert_called_once_with(
            "/etc/modprobe.d/ttm.conf", "w", encoding="utf-8"
        )
        mock_print.assert_any_call(
            "Successfully set TTM pages limit to 20480 pages (5.00 GB)", "🐧"
        )
        self.assertTrue(result)

    @mock.patch("os.path.exists", return_value=False)
    @mock.patch("amd_debug.ttm.print_color")
    def test_clear_file_not_exists(self, mock_print, _mock_exists):
        """Test clear() when config file does not exist"""
        result = self.tool.clear()
        mock_print.assert_called_with("/etc/modprobe.d/ttm.conf doesn't exist", "❌")
        self.assertFalse(result)

    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("amd_debug.ttm.relaunch_sudo", return_value=True)
    @mock.patch("os.remove")
    @mock.patch("amd_debug.ttm.print_color")
    @mock.patch("amd_debug.ttm.maybe_reboot", return_value=True)
    def test_clear_success(
        self, _mock_reboot, mock_print, mock_remove, _mock_relaunch_sudo, _mock_exists
    ):
        """Test clear() success path"""
        result = self.tool.clear()
        mock_remove.assert_called_once_with("/etc/modprobe.d/ttm.conf")
        mock_print.assert_any_call(
            "Configuration /etc/modprobe.d/ttm.conf removed", "🐧"
        )
        self.assertTrue(result)



# SPDX-License-Identifier: MIT


def amd_s2idle():
    """Launch the amd-s2idle tool."""
    from . import s2idle  # pylint: disable=import-outside-toplevel

    return s2idle.main()


def amd_bios():
    """Launch the amd-bios tool."""
    from . import bios  # pylint: disable=import-outside-toplevel

    return bios.main()


def amd_pstate():
    """Launch the amd-pstate tool."""
    from . import pstate  # pylint: disable=import-outside-toplevel

    return pstate.main()


def amd_ttm():
    """Launch the amd-ttm tool."""
    from . import ttm  # pylint: disable=import-outside-toplevel

    return ttm.main()


def install_dep_superset():
    """Install all superset dependencies."""
    from . import installer  # pylint: disable=import-outside-toplevel

    return installer.install_dep_superset()


def launch_tool(tool_name):
    """Launch a tool with the given name and arguments."""
    tools = {
        "amd_s2idle.py": amd_s2idle,
        "amd_bios.py": amd_bios,
        "amd_pstate.py": amd_pstate,
        "amd_ttm.py": amd_ttm,
        "install_deps.py": install_dep_superset,
    }
    if tool_name in tools:
        return tools[tool_name]()
    else:
        print(f"\033[91mUnknown exe: {tool_name}\033[0m")
        return 1



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the installer functions in the amd-debug-tools package.
"""
from unittest.mock import patch, mock_open, MagicMock

import logging
import subprocess
import sys
import unittest

from amd_debug.installer import (
    Installer,
    DistroPackage,
    FwupdPackage,
    EdidDecodePackage,
    install_dep_superset,
    parse_args,
    show_install_message,
)


class TestInstaller(unittest.TestCase):
    """Test installer functions"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    def setUp(self):
        self.installer = Installer(tool_debug=False)

    @patch("builtins.print")
    @patch("shutil.copy", return_value=None)
    @patch("os.chmod", return_value=None)
    @patch("builtins.open")
    @patch("subprocess.call", return_value=0)
    @patch("os.makedirs", return_value=None)
    def test_install_hook(
        self,
        _mock_mkdir,
        _mock_call,
        _mock_open,
        _mock_chmod,
        _mock_shutil,
        _mock_print,
    ):
        """Test install hook function"""
        self.installer.install()

    @patch("builtins.print")
    @patch("shutil.copy", return_value=None)
    @patch("os.chmod", return_value=None)
    @patch("builtins.open")
    @patch("subprocess.call", return_value=0)
    @patch("os.makedirs", return_value=None)
    def test_install_hook_missing_path(
        self,
        _mock_mkdir,
        _mock_call,
        _mock_open,
        _mock_chmod,
        _mock_shutil,
        _mock_print,
    ):
        """Test install hook function, but some paths are missing"""
        with patch("os.path.exists", return_value=False):
            self.installer.install()

    @patch("builtins.print")
    @patch("os.remove", return_value=None)
    @patch("builtins.open")
    @patch("amd_debug.installer.get_distro", return_value="ubuntu")
    @patch("subprocess.call", return_value=0)
    def test_remove_hook(
        self, _mock_call, _mock_get_distro, _mock_open, _mock_remove, _mock_print
    ):
        """Test remove hook function"""
        with patch("os.path.exists", return_value=True):
            self.installer.remove()

    @patch("builtins.print")
    @patch("os.path.exists", return_value=False)
    @patch("subprocess.call", return_value=0)
    def test_remove_hook_missing_path(self, _mock_call, _mock_exists, _mock_print):
        """Test remove hook function when the file is missing"""
        self.installer.remove()

    @patch("builtins.print")
    @patch("subprocess.call", return_value=0)
    def test_already_installed_iasl(self, _mock_call, _mock_print):
        """Test that an already installed iasl is found"""
        self.installer.set_requirements("iasl")
        ret = self.installer.install_dependencies()
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="ubuntu")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_iasl_ubuntu(
        self, _mock_call, _mock_check_call, _mock_distro, _fake_sudo, _mock_print
    ):
        """Test install requirements function"""
        self.installer.set_requirements("iasl")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["apt", "install", "acpica-tools"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="VARIANT_ID=workstation\n"
    )
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_iasl_fedora(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function"""
        self.installer.set_requirements("iasl")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(
            ["dnf", "install", "-y", "acpica-tools"]
        )
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch("builtins.open", new_callable=mock_open, read_data="VARIANT_ID=kde\n")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_iasl_fedora_kde(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function on Fedora KDE"""
        self.installer.set_requirements("iasl")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(
            ["dnf", "install", "-y", "acpica-tools"]
        )
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="ubuntu")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_ethtool_ubuntu(
        self, _mock_call, _mock_check_call, _mock_distro, _fake_sudo, _mock_print
    ):
        """Test install requirements function"""
        self.installer.set_requirements("ethtool")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["apt", "install", "ethtool"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="VARIANT_ID=workstation\n"
    )
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_ethtool_fedora(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function"""
        self.installer.set_requirements("ethtool")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["dnf", "install", "-y", "ethtool"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch("builtins.open", new_callable=mock_open, read_data="VARIANT_ID=kde\n")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_ethtool_fedora_kde(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function on Fedora KDE"""
        self.installer.set_requirements("ethtool")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["dnf", "install", "-y", "ethtool"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="arch")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_ethtool_arch(
        self,
        _mock_call,
        _mock_check_call,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function"""
        self.installer.set_requirements("ethtool")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["pacman", "-Sy", "ethtool"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("os.path.exists", return_value=False)
    @patch("os.execvp", return_value=None)
    @patch("amd_debug.installer.get_distro", return_value="gentoo")
    @patch("subprocess.call", return_value=1)
    def test_install_iasl_gentoo(
        self, _mock_call, _mock_distro, _fake_sudo, _mock_exists, _mock_print
    ):
        """Test install requirements function"""
        self.installer.set_requirements("iasl", "ethtool")
        ret = self.installer.install_dependencies()
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="ubuntu")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_edid_decode_ubuntu(
        self, _mock_call, _mock_check_call, _mock_distro, _fake_sudo, _mock_print
    ):
        """Test install requirements function for edid-decode on Ubuntu"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(
            ["apt", "install", "libdisplay-info-bin"]
        )
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch(
        "builtins.open", new_callable=mock_open, read_data="VARIANT_ID=workstation\n"
    )
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_edid_decode_fedora(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function for edid-decode on Fedora"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(
            ["dnf", "install", "-y", "libdisplay-info-tools"]
        )
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch("builtins.open", new_callable=mock_open, read_data="VARIANT_ID=kde\n")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_edid_decode_fedora_kde(
        self,
        _mock_call,
        _mock_check_call,
        _mock_variant,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function for edid-decode on Fedora KDE"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(
            ["dnf", "install", "-y", "libdisplay-info-tools"]
        )
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("amd_debug.installer.get_distro", return_value="arch")
    @patch("os.execvp", return_value=None)
    @patch("subprocess.check_call", return_value=0)
    @patch("subprocess.call", return_value=1)
    def test_install_edid_decode_arch(
        self,
        _mock_call,
        _mock_check_call,
        _mock_distro,
        _fake_sudo,
        _mock_print,
    ):
        """Test install requirements function for edid-decode on Arch"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        _mock_check_call.assert_called_once_with(["pacman", "-Sy", "libdisplay-info"])
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("os.path.exists", return_value=False)
    @patch("os.execvp", return_value=None)
    @patch("amd_debug.installer.get_distro", return_value="gentoo")
    @patch("subprocess.call", return_value=1)
    def test_install_edid_decode_gentoo(
        self, _mock_call, _mock_distro, _fake_sudo, _mock_exists, _mock_print
    ):
        """Test install requirements function for edid-decode on unsupported distro"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("os.path.exists", return_value=False)
    @patch("os.execvp", return_value=None)
    @patch("amd_debug.installer.get_distro", return_value="gentoo")
    @patch("subprocess.call", return_value=255)
    def test_install_edid_decode_present(
        self, _mock_call, _mock_distro, _fake_sudo, _mock_exists, _mock_print
    ):
        """Test install requirements function for edid-decode on unsupported distro"""
        self.installer.set_requirements("edid-decode")
        ret = self.installer.install_dependencies()
        self.assertTrue(ret)

    @patch("builtins.print")
    @patch("subprocess.call", side_effect=FileNotFoundError())
    @patch("amd_debug.installer.get_distro", return_value="gentoo")
    @patch("os.path.exists", return_value=False)
    @patch("os.execvp", return_value=None)
    def test_install_iasl_missing_filenotfound(
        self, _execvp, _exists, _distro, _call, _print
    ):
        """When iasl binary is missing entirely (FileNotFoundError) installer still tries"""
        self.installer.set_requirements("iasl")
        self.assertTrue(self.installer.install_dependencies())

    @patch("builtins.print")
    @patch("subprocess.call", side_effect=FileNotFoundError())
    @patch("amd_debug.installer.get_distro", return_value="gentoo")
    @patch("os.path.exists", return_value=False)
    @patch("os.execvp", return_value=None)
    def test_install_ethtool_missing_filenotfound(
        self, _execvp, _exists, _distro, _call, _print
    ):
        """When ethtool binary is missing (FileNotFoundError) installer still tries"""
        self.installer.set_requirements("ethtool")
        self.assertTrue(self.installer.install_dependencies())

    @patch("builtins.print")
    @patch("subprocess.call", side_effect=FileNotFoundError())
    @patch("amd_debug.installer.get_distro", return_value="gentoo")
    @patch("os.path.exists", return_value=False)
    @patch("os.execvp", return_value=None)
    def test_install_edid_decode_both_missing(
        self, _execvp, _exists, _distro, _call, _print
    ):
        """Both di-edid-decode and edid-decode missing; falls into install path"""
        self.installer.set_requirements("edid-decode")
        self.assertTrue(self.installer.install_dependencies())

    @patch("builtins.print")
    @patch("os.execvp", return_value=None)
    def test_install_pyudev_present(self, _execvp, _print):
        """pyudev already importable; no install needed"""
        self.installer.set_requirements("pyudev")
        self.assertTrue(self.installer.install_dependencies())

    @patch("builtins.print")
    @patch("os.execvp", return_value=None)
    def test_install_packaging_present(self, _execvp, _print):
        """packaging already importable"""
        self.installer.set_requirements("packaging")
        self.assertTrue(self.installer.install_dependencies())

    @patch("builtins.print")
    @patch("os.execvp", return_value=None)
    def test_install_pandas_present(self, _execvp, _print):
        """pandas already importable"""
        self.installer.set_requirements("pandas")
        self.assertTrue(self.installer.install_dependencies())

    @patch("builtins.print")
    @patch("os.execvp", return_value=None)
    def test_install_tabulate_present(self, _execvp, _print):
        """tabulate already importable"""
        self.installer.set_requirements("tabulate")
        self.assertTrue(self.installer.install_dependencies())

    @patch("builtins.print")
    @patch("os.execvp", return_value=None)
    def test_install_jinja2_present(self, _execvp, _print):
        """jinja2 already importable"""
        self.installer.set_requirements("jinja2")
        self.assertTrue(self.installer.install_dependencies())

    @patch("builtins.print")
    @patch("os.execvp", return_value=None)
    def test_install_seaborn_present(self, _execvp, _print):
        """seaborn already importable"""
        self.installer.set_requirements("seaborn")
        self.assertTrue(self.installer.install_dependencies())

    @patch("amd_debug.installer.print_color")
    @patch("amd_debug.installer.get_distro", return_value="ubuntu")
    @patch("amd_debug.installer.relaunch_sudo")
    def test_distro_package_no_deb_returns_false(
        self, _sudo, _distro, _print
    ):
        """Ubuntu install path returns False when deb name is missing"""
        pkg = DistroPackage(deb=None, rpm="x", arch="x", message="msg")
        self.assertFalse(pkg.install())

    @patch("amd_debug.installer.print_color")
    @patch("amd_debug.installer.read_file", return_value="VARIANT_ID=workstation\n")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch("amd_debug.installer.relaunch_sudo")
    def test_distro_package_no_rpm_returns_false(
        self, _sudo, _distro, _read, _print
    ):
        """Fedora install path returns False when rpm name is missing"""
        pkg = DistroPackage(deb="x", rpm=None, arch="x", message="msg")
        self.assertFalse(pkg.install())

    @patch("amd_debug.installer.print_color")
    @patch("amd_debug.installer.read_file", return_value="VARIANT_ID=server\n")
    @patch("amd_debug.installer.get_distro", return_value="fedora")
    @patch("amd_debug.installer.relaunch_sudo")
    def test_distro_package_fedora_wrong_variant(
        self, _sudo, _distro, _read, _print
    ):
        """Fedora install path returns False when variant isn't workstation/kde"""
        pkg = DistroPackage(deb="x", rpm="x", arch="x", message="msg")
        self.assertFalse(pkg.install())

    @patch("amd_debug.installer.print_color")
    @patch("amd_debug.installer.get_distro", return_value="arch")
    @patch("amd_debug.installer.relaunch_sudo")
    def test_distro_package_no_arch_returns_false(self, _sudo, _distro, _print):
        """Arch install path returns False when arch name is missing"""
        pkg = DistroPackage(deb="x", rpm="x", arch=None, message="msg")
        self.assertFalse(pkg.install())

    @patch("amd_debug.installer.print_color")
    @patch("amd_debug.installer.fatal_error")
    @patch(
        "amd_debug.installer.subprocess.check_call",
        side_effect=subprocess.CalledProcessError(1, "apt"),
    )
    @patch("amd_debug.installer.get_distro", return_value="ubuntu")
    @patch("amd_debug.installer.relaunch_sudo")
    def test_distro_package_install_failure_calls_fatal(
        self, _sudo, _distro, _call, mock_fatal, _print
    ):
        """check_call failure triggers fatal_error"""
        pkg = DistroPackage(deb="x", rpm="x", arch="x", message="msg")
        self.assertTrue(pkg.install())
        mock_fatal.assert_called_once()

    def test_show_install_message(self):
        """show_install_message formats and prints"""
        with patch("amd_debug.installer.print_color") as mock_pc:
            show_install_message("Thing missing")
            mock_pc.assert_called_once()
            self.assertIn("Thing missing", mock_pc.call_args[0][0])
            self.assertIn("Attempting to install", mock_pc.call_args[0][0])

    def test_parse_args_tool_debug(self):
        """parse_args parses --tool-debug"""
        with patch("sys.argv", ["installer", "--tool-debug"]):
            args = parse_args()
        self.assertTrue(args.tool_debug)

    @patch("amd_debug.installer.show_log_info")
    @patch("amd_debug.installer.Installer")
    @patch("amd_debug.installer.parse_args")
    def test_install_dep_superset_success(
        self, mock_parse, mock_installer_cls, _mock_show
    ):
        """install_dep_superset returns None on success"""
        mock_parse.return_value = MagicMock(tool_debug=False)
        mock_tool = MagicMock()
        mock_tool.install_dependencies.return_value = True
        mock_installer_cls.return_value = mock_tool
        self.assertIsNone(install_dep_superset())
        mock_tool.set_requirements.assert_called_once()

    @patch("amd_debug.installer.show_log_info")
    @patch("amd_debug.installer.Installer")
    @patch("amd_debug.installer.parse_args")
    def test_install_dep_superset_failure(
        self, mock_parse, mock_installer_cls, _mock_show
    ):
        """install_dep_superset returns 1 when dependencies fail to install"""
        mock_parse.return_value = MagicMock(tool_debug=False)
        mock_tool = MagicMock()
        mock_tool.install_dependencies.return_value = False
        mock_installer_cls.return_value = mock_tool
        self.assertEqual(install_dep_superset(), 1)

    def test_installer_fwupd_import_error(self):
        """Installer sets fwupd=False when gi.repository.Fwupd import fails"""
        # Simulate ImportError when importing gi by removing/setting None
        with patch.dict(sys.modules, {"gi": None}):
            tool = Installer(tool_debug=False)
            self.assertFalse(tool.fwupd)



#!/usr/bin/python3
# SPDX-License-Identifier: MIT

"""
This module contains unit tests for the pstate tool in the amd-debug-tools package.
"""

import logging
import unittest
from unittest.mock import patch, MagicMock

from amd_debug.pstate import (
    AmdPstateTriage,
    amd_cppc_cap_lowest_perf,
    amd_cppc_cap_lownonlin_perf,
    amd_cppc_cap_nominal_perf,
    amd_cppc_cap_highest_perf,
    amd_cppc_max_perf,
    amd_cppc_min_perf,
    amd_cppc_des_perf,
    amd_cppc_epp_perf,
)


class TestAmdPstateTriage(unittest.TestCase):
    """Test AmdPstateTriage class"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    @patch("amd_debug.pstate.relaunch_sudo")
    @patch("amd_debug.pstate.get_pretty_distro", return_value="Test Distro")
    @patch("amd_debug.pstate.print_color")
    @patch("amd_debug.pstate.Context")
    def test_init(
        self,
        _mock_context,
        mock_print_color,
        mock_get_pretty_distro,
        mock_relaunch_sudo,
    ):
        """Test initialization of AmdPstateTriage class"""
        triage = AmdPstateTriage(logging=True)
        mock_relaunch_sudo.assert_called_once()
        mock_get_pretty_distro.assert_called_once()
        mock_print_color.assert_called_with("Test Distro", "🐧")
        self.assertIsNotNone(triage.context)

    @patch("amd_debug.pstate.os.path.exists", return_value=True)
    @patch("amd_debug.pstate.read_file", return_value="test_value")
    @patch("amd_debug.pstate.print_color")
    @patch("amd_debug.pstate.relaunch_sudo")
    def test_gather_amd_pstate_info(
        self, _mock_relaunch_sudo, mock_print_color, mock_read_file, mock_path_exists
    ):
        """Test gather_amd_pstate_info method"""
        triage = AmdPstateTriage(logging=False)
        triage.gather_amd_pstate_info()
        mock_path_exists.assert_called()
        mock_read_file.assert_called()
        mock_print_color.assert_any_call("'status':\ttest_value", "○")
        mock_print_color.assert_any_call("'prefcore':\ttest_value", "○")

    @patch(
        "amd_debug.pstate.os.uname",
        return_value=MagicMock(sysname="Linux", release="5.15.0"),
    )
    @patch("amd_debug.pstate.print_color")
    @patch("amd_debug.pstate.relaunch_sudo")
    def test_gather_kernel_info(
        self, _mock_relaunch_sudo, mock_print_color, mock_uname
    ):
        """Test gather_kernel_info method"""
        triage = AmdPstateTriage(logging=False)
        triage.gather_kernel_info()
        mock_uname.assert_called()
        mock_print_color.assert_called_with("Kernel:\t5.15.0", "🐧")

    @patch("amd_debug.pstate.os.path.exists", return_value=True)
    @patch("amd_debug.pstate.read_file", return_value="1")
    @patch("amd_debug.pstate.print_color")
    @patch("amd_debug.pstate.relaunch_sudo")
    def test_gather_scheduler_info(
        self, _mock_relaunch_sudo, mock_print_color, mock_read_file, mock_path_exists
    ):
        """Test gather_scheduler_info method"""
        triage = AmdPstateTriage(logging=False)
        triage.gather_scheduler_info()
        mock_path_exists.assert_called()
        mock_read_file.assert_called()
        mock_print_color.assert_any_call("ITMT:\t1", "🐧")

    @patch("amd_debug.pstate.read_msr")
    @patch("amd_debug.pstate.print_color")
    @patch("amd_debug.pstate.Context")
    @patch("amd_debug.pstate.relaunch_sudo")
    def test_gather_msrs(
        self, _mock_relaunch_sudo, mock_context, mock_print_color, mock_read_msr
    ):
        """Test gather_msrs method"""
        # Mock the list of CPUs
        mock_context.return_value.list_devices.return_value = [
            MagicMock(sys_name="cpu0"),
            MagicMock(sys_name="cpu1"),
        ]

        # Mock MSR values for the CPUs
        mock_read_msr.side_effect = [
            0x1,  # MSR_AMD_CPPC_ENABLE for cpu0
            0x2,  # MSR_AMD_CPPC_STATUS for cpu0
            0x12345678,  # MSR_AMD_CPPC_CAP1 for cpu0
            0x87654321,  # MSR_AMD_CPPC_CAP2 for cpu0
            0xABCDEF,  # MSR_AMD_CPPC_REQ for cpu0
            0x1,  # MSR_AMD_CPPC_ENABLE for cpu1
            0x2,  # MSR_AMD_CPPC_STATUS for cpu1
            0x12345678,  # MSR_AMD_CPPC_CAP1 for cpu1
            0x87654321,  # MSR_AMD_CPPC_CAP2 for cpu1
            0xABCDEF,  # MSR_AMD_CPPC_REQ for cpu1
        ]

        triage = AmdPstateTriage(logging=False)
        result = triage.gather_msrs()

        # Assert that MSR values were read for both CPUs
        self.assertEqual(mock_read_msr.call_count, 10)

        # Assert that print_color was called to display the MSR information
        self.assertTrue(mock_print_color.called)

        # Assert that the method returned True (indicating success)
        self.assertIsNone(result)

    def test_amd_cppc_cap_lowest_perf(self):
        """Test amd_cppc_cap_lowest_perf function"""
        self.assertEqual(amd_cppc_cap_lowest_perf(0x12345678), 0x78)

    def test_amd_cppc_cap_lownonlin_perf(self):
        """Test amd_cppc_cap_lownonlin_perf function"""
        self.assertEqual(amd_cppc_cap_lownonlin_perf(0x12345678), 0x56)

    def test_amd_cppc_cap_nominal_perf(self):
        """Test amd_cppc_cap_nominal_perf function"""
        self.assertEqual(amd_cppc_cap_nominal_perf(0x12345678), 0x34)

    def test_amd_cppc_cap_highest_perf(self):
        """Test amd_cppc_cap_highest_perf function"""
        self.assertEqual(amd_cppc_cap_highest_perf(0x12345678), 0x12)

    def test_amd_cppc_max_perf(self):
        """Test amd_cppc_max_perf function"""
        self.assertEqual(amd_cppc_max_perf(0x12345678), 0x78)

    def test_amd_cppc_min_perf(self):
        """Test amd_cppc_min_perf function"""
        self.assertEqual(amd_cppc_min_perf(0x12345678), 0x56)

    def test_amd_cppc_des_perf(self):
        """Test amd_cppc_des_perf function"""
        self.assertEqual(amd_cppc_des_perf(0x12345678), 0x34)

    def test_amd_cppc_epp_perf(self):
        """Test amd_cppc_epp_perf function"""
        self.assertEqual(amd_cppc_epp_perf(0x12345678), 0x12)

    @patch("amd_debug.pstate.read_file", return_value="data")
    @patch("amd_debug.pstate.os.path.exists", return_value=True)
    @patch("amd_debug.pstate.print_color")
    @patch("amd_debug.pstate.Context")
    @patch("amd_debug.pstate.relaunch_sudo")
    def test_gather_cpu_info(
        self,
        _mock_relaunch_sudo,
        mock_context,
        mock_print_color,
        _mock_path_exists,
        mock_read_file,
    ):
        """gather_cpu_info builds a CPU dataframe and prints it"""
        cpu0 = MagicMock(sys_name="cpu0", sys_path="/sys/devices/system/cpu/cpu0")
        cpu1 = MagicMock(sys_name="cpu1", sys_path="/sys/devices/system/cpu/cpu1")
        mock_context.return_value.list_devices.return_value = [cpu1, cpu0]

        # cpuinfo read needs to contain a "model name" line
        def _read(path):
            if path == "/proc/cpuinfo":
                return "model name\t: AMD Test CPU\n"
            return "1"

        mock_read_file.side_effect = _read

        triage = AmdPstateTriage(logging=False)
        triage.gather_cpu_info()
        # ensure CPU model line printed
        mock_print_color.assert_any_call("CPU:\t\tAMD Test CPU", "💻")

    @patch("amd_debug.pstate.read_msr", side_effect=FileNotFoundError())
    @patch("amd_debug.pstate.print_color")
    @patch("amd_debug.pstate.Context")
    @patch("amd_debug.pstate.relaunch_sudo")
    def test_gather_msrs_module_missing(
        self, _mock_relaunch_sudo, mock_context, mock_print_color, _mock_read_msr
    ):
        """gather_msrs returns False when MSR module not loaded"""
        mock_context.return_value.list_devices.return_value = [
            MagicMock(sys_name="cpu0"),
        ]
        triage = AmdPstateTriage(logging=False)
        result = triage.gather_msrs()
        self.assertFalse(result)
        mock_print_color.assert_any_call(
            "Unable to check MSRs: MSR kernel module not loaded", "❌"
        )

    @patch("amd_debug.pstate.read_msr", side_effect=PermissionError())
    @patch("amd_debug.pstate.print_color")
    @patch("amd_debug.pstate.Context")
    @patch("amd_debug.pstate.relaunch_sudo")
    def test_gather_msrs_permission(
        self, _mock_relaunch_sudo, mock_context, mock_print_color, _mock_read_msr
    ):
        """gather_msrs returns silently when MSR reads denied"""
        mock_context.return_value.list_devices.return_value = [
            MagicMock(sys_name="cpu0"),
        ]
        triage = AmdPstateTriage(logging=False)
        result = triage.gather_msrs()
        self.assertIsNone(result)
        mock_print_color.assert_any_call("MSR checks unavailable", "🚦")

    @patch("amd_debug.pstate.print_color")
    @patch("amd_debug.pstate.Context")
    @patch("amd_debug.pstate.relaunch_sudo")
    def test_run_cpu_info_failure(
        self, _mock_relaunch_sudo, _mock_context, mock_print_color
    ):
        """run returns False when gather_cpu_info raises FileNotFoundError"""
        triage = AmdPstateTriage(logging=False)
        with patch.object(triage, "gather_kernel_info"), \
             patch.object(triage, "gather_amd_pstate_info"), \
             patch.object(triage, "gather_scheduler_info"), \
             patch.object(
                 triage, "gather_cpu_info", side_effect=FileNotFoundError()
             ):
            result = triage.run()
        self.assertFalse(result)
        mock_print_color.assert_any_call("Unable to gather CPU information", "❌")

    @patch("amd_debug.pstate.print_color")
    @patch("amd_debug.pstate.Context")
    @patch("amd_debug.pstate.relaunch_sudo")
    def test_run_success(self, _mock_relaunch_sudo, _mock_context, _mock_print_color):
        """run returns True when all gather steps succeed"""
        triage = AmdPstateTriage(logging=False)
        with patch.object(triage, "gather_kernel_info"), \
             patch.object(triage, "gather_amd_pstate_info"), \
             patch.object(triage, "gather_scheduler_info"), \
             patch.object(triage, "gather_cpu_info"), \
             patch.object(triage, "gather_msrs"):
            self.assertTrue(triage.run())


class TestPstateMain(unittest.TestCase):
    """Test the pstate module-level main/parse_args helpers"""

    @classmethod
    def setUpClass(cls):
        logging.basicConfig(filename="/dev/null", level=logging.DEBUG)

    @patch("amd_debug.pstate.sys.argv", ["amd_pstate"])
    def test_parse_args_no_args_exits(self):
        """parse_args exits when no arguments provided"""
        from amd_debug.pstate import parse_args
        with self.assertRaises(SystemExit):
            parse_args()

    @patch("amd_debug.pstate.sys.argv", ["amd_pstate", "--version"])
    def test_parse_args_version(self):
        """parse_args parses --version"""
        from amd_debug.pstate import parse_args
        args = parse_args()
        self.assertTrue(args.version)

    @patch("amd_debug.pstate.sys.argv", ["amd_pstate", "triage", "--tool-debug"])
    def test_parse_args_triage(self):
        """parse_args parses triage subcommand"""
        from amd_debug.pstate import parse_args
        args = parse_args()
        self.assertEqual(args.command, "triage")
        self.assertTrue(args.tool_debug)

    @patch("amd_debug.pstate.version", return_value="1.2.3")
    @patch("amd_debug.pstate.parse_args")
    def test_main_version(self, mock_parse, mock_version):
        """main prints version and returns None"""
        from amd_debug.pstate import main
        mock_parse.return_value = MagicMock(version=True, command=None)
        self.assertIsNone(main())
        mock_version.assert_called_once()

    @patch("amd_debug.pstate.show_log_info")
    @patch("amd_debug.pstate.AmdPstateTriage")
    @patch("amd_debug.pstate.parse_args")
    def test_main_triage_success(self, mock_parse, mock_triage_cls, _mock_show):
        """main returns None when triage succeeds"""
        from amd_debug.pstate import main
        mock_parse.return_value = MagicMock(
            version=False, command="triage", tool_debug=False
        )
        mock_triage_cls.return_value.run.return_value = True
        self.assertIsNone(main())

    @patch("amd_debug.pstate.show_log_info")
    @patch("amd_debug.pstate.AmdPstateTriage")
    @patch("amd_debug.pstate.parse_args")
    def test_main_triage_failure(self, mock_parse, mock_triage_cls, _mock_show):
        """main returns 1 when triage fails"""
        from amd_debug.pstate import main
        mock_parse.return_value = MagicMock(
            version=False, command="triage", tool_debug=False
        )
        mock_triage_cls.return_value.run.return_value = False
        self.assertEqual(main(), 1)
